@@ -3,6 +3,7 @@
 import asyncio
 import logging
 from collections import deque
+from pprint import pprint  # noqa: W0611
 from threading import Barrier, Thread
 
 import janus
@@ -61,9 +62,9 @@ class Node(Component):
         await super().run()
         for layer in self.layers:
             if isinstance(layer, Component) and hasattr(layer, 'ready'):
-                self.info(f'Waiting on layer {layer} to be ready')
+                self.debug(f'Waiting on layer {layer} to be ready')
                 await layer.ready.wait()
-                self.info(f'Layer {layer} is ready')
+                self.debug(f'Layer {layer} is ready')
         # TODO: handle exceptions, and possibly restart
         if self.running and self.inbox.running:
             stack = self.layers[0]
@@ -83,6 +84,7 @@ class Node(Component):
         processor = self.get_processor()
         if self.blocking:
             raise NotImplementedError(f'Node.blocking in {self}')
+            # processor = ThreadedProcessor(processor, scale=self.scale)
 
         if self.scale == 1:
             return processor
@@ -114,6 +116,7 @@ class Node(Component):
     process.stub = True
 
     def drop(self, message):  # TODO: implement message accounting and leak detection
+        # self.outbox.drop(message)
         raise NotImplementedError(f'drop() in {self}')
 
     def merged_settings(self, message, key=None, msgmap=None):
@@ -145,20 +148,30 @@ class MultiProcessor:
 
     async def feeder(self, messages):
         """Loop through all messages from the inbox and relay them to the input queue."""
+        # self.node.debug(f'*** collector feeder running')
         async for channel, message in messages:
             self.pending += 1
+            # self.node.debug(f'*** collector feeder got msg [{self.pending}] {message}')
             await self.iqueue.put((channel, message))
+            # self.node.debug(f'*** collector feeder sent msg [{self.pending}] {message}')
+        # self.node.debug(f'*** collector feeder done')
         self.exhausted = True
         if self.pending > 0:
             await self.finished.wait()
         for i in range(self.node.scale):
             await self.iqueue.put(Message.EOT)
+        # self.node.debug(f'*** collector feeder exited')
 
     async def collector(self, id_):
         """Launch an instance of the node processor, and relay its output to the output queue."""
+        # self.node.debug(f'*** collector [{id_}] running')
         async for channel, message in self.processor(aiter(self.iqueue.get, Message.EOT), id_=id_):
+            # self.node.debug(f'*** collector [{id_}] got msg {message}')
             await self.oqueue.put((channel, message))
+            # self.node.debug(f'*** collector [{id_}] sent msg {message}')
+        # self.node.debug(f'*** collector [{id_}] stopped')
         await self.oqueue.put((None, Message.EOT))
+        # self.node.debug(f'*** collector [{id_}] exited')
 
     async def outfeed(self, messages):
         """Launch all processor instances and return an iterator that feeds on the output queue."""
@@ -171,19 +184,27 @@ class MultiProcessor:
         for i in range(self.node.scale):
             coroutines.append(self.node.loop.create_task(self.collector(i)))
         running = self.node.scale
+        # self.node.debug(f'*** started {running} collectors')
         while running or not self.oqueue.empty():
             channel, message = await self.oqueue.get()
+            # self.node.debug(f'*** collector outfeed got msg [{self.pending}] {message}')
             self.oqueue.task_done()
             if message is Message.EOT:
                 running -= 1
+                # self.node.debug(f'*** a collector ended, left={running}')
             else:
                 self.pending -= 1
                 if self.pending == 0 and self.exhausted:
+                    # self.node.debug(f'*** collector finishing')
                     self.finished.set()
                 yield channel, message
 
+            # self.node.debug(f'*** collector outfeed processed msg [{self.pending}] {message}')
+
+        # self.node.debug(f'*** waiting on all collectors')
         for coroutine in coroutines:
             await coroutine
+        # self.node.debug(f'*** all collectors ended')
 
     __call__ = outfeed
 
@@ -200,6 +221,7 @@ class Inbox(Manifold):
                 return
 
             try:
+                # self.debug(f'* waiting for message on channel {channel}')
                 message = await pipe.receive()
             except Exception:
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
@@ -225,10 +247,60 @@ class Inbox(Manifold):
                 else:
                     yield channel, pipe
 
+    async def Xreceiver(self):
+        self.debug('Receiving')
+        self.debug(f'Setting up channels {list(self.channels)}')
+        # __import__('pudb').set_trace()
+        # import sys; sys.exit()
+        feeder = deque()
+        events = Box(received=asyncio.Event())
+        # receivers = dict()
+        for channel, pipe in self._active_channels():
+            self.debug(f'Setting up channel {channel}')
+            events[channel] = asyncio.Event()
+            events[channel].set()
+            # receivers[channel] = asyncio.ensure_future(
+            #     self.channel_receiver(channel, feeder, events))
+            asyncio.ensure_future(self.channel_receiver(channel, feeder, events))
+
+        self.debug('Waiting for feeder')
+        while self.running and (len(feeder) or await events.received.wait()):
+            for channel in self.channels:
+                events[channel].clear()
+            events.received.clear()
+            channel, message = feeder.popleft()
+
+            self.debug(f'Got message from {channel}: {message}')
+            if message == Message.EOT:
+                self.debug('Exhausted channel: %s' % channel)
+                pipe.stop()
+                # self.debug(f'* setting event for channel {channel}')
+                events[channel].set()
+                # receivers[channel].cancel()
+                try:
+                    await pipe.send(Message.EOT)
+                except NotImplementedError:
+                    pass
+                # self.debug('-----------------------------')
+                if not list(self._active_channels()):
+                    self.debug('Stopping: no more active channels')
+                    self.stop()
+                    # TODO: cancel channel receiver coroutines
+                    break
+            else:
+                message.checkin(self)
+                yield channel, message
+
+            for channel in self.channels:
+                events[channel].set()
+
+        self.debug('Finished receiving')
+
     async def channel_receiver(self, channel, feeder, events):
         self.debug(f'Receiver for {channel} is ready')
         pipe = self.channels[channel]
         while pipe.running and self.running and await events[channel].wait():
+            # self.debug(f'* receiving from channel {channel}')
             events[channel].clear()
             try:
                 feeder.append((channel, await pipe.receive()))
@@ -236,9 +308,22 @@ class Inbox(Manifold):
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
                 raise
 
+            # self.debug(f'* received from channel {channel} (next event: {events[channel].is_set()})')
             events.received.set()
         self.debug(f'Receiver for {channel} is exiting')
 
+    def X_active_channels(self):
+        channels = self.channels.copy()
+        while channels:
+            for channel, pipe in channels.copy().items():
+                del channels[channel]
+                if not pipe.running:
+                    self.debug(f'skipping stopped channel {channel}')
+                else:
+                    yield channel, pipe
+        if channels:
+            pass  # TODO: channel clean-ups
+
 
 class RetryableMessages:
     """Interfaces a queue between the inbox and the processor where we can insert back retries."""
@@ -256,13 +341,17 @@ class RetryableMessages:
     async def feeder(self):
         """Feed the inbox to the queue."""
         async for channel, self.newmsg in self.messages:
+            # self.node.debug(f'*** collector retry: feeding {self.newmsg.type} msg {self.newmsg.meta.id}')
             self.forwarded.clear()
             await self.queue.put((channel, self.newmsg))
             await self.forwarded.wait()
+        # self.node.debug(f'*** collector retry: feeder done')
         self.exhausted = True
         if self.pending > 0:
             await self.finished.wait()
+        # self.node.debug(f'*** collector retry: feeder ready to exit')
         await self.queue.put(Message.EOT)
+        # self.node.debug(f'*** collector retry: feeder exited')
 
     async def forwarder(self):
         """Turn the queue into an iterator that the processor can consume."""
@@ -274,23 +363,28 @@ class RetryableMessages:
         self.pending = 0  # number of messages currently handled by node processor
         feeder = self.node.loop.create_task(self.feeder())
         async for channel, message in aiter(self.queue.get, Message.EOT):
+            # self.node.debug(f'*** collector retry: forwarding {message.type} msg {message.meta.id}')
             self.queue.task_done()
             if message is self.newmsg:
                 self.forwarded.set()
             self.pending += 1
             yield channel, message
 
+        # self.node.debug(f'*** collector retry: forwarder exiting')
         self.queue.task_done()
         await feeder
+        # self.node.debug(f'*** collector retry: forwarder exited')
 
     def processing_done(self):
         """Signal that a message is not in the processor anymore."""
         self.pending -= 1
+        # self.node.debug(f'*** collector retry: processing done {self.pending}')
         if self.exhausted:
             self.finished.set()
 
     async def retry(self, channel, message, wait=0.001, id_=None):
         """Insert a message back into the queue for retrying it."""
+        # self.node.debug(f'*** collector retry{"" if id_ is None else " %s" % id_}: retrying {message.type} msg {message.meta.id}')
         await asyncio.sleep(wait)
         await self.node.loop.create_task(self.queue.put((channel, message)))
         self.processing_done()
@@ -310,9 +404,11 @@ class Outbox(Manifold):
                 self.debug(f'Sending to {channel}: {message}')
                 if channel in self.channels:
                     await self.channels[channel].send(message)
+                    # self.debug(f'Sent to    {channel}: {message} [{self.channels[channel].id}]')
                     message.checkout(self)
                     yield channel, message
 
+                    # self.debug(f'Yielded    {channel}: {message}')
                 else:
                     raise KeyError(f'Channel "{channel}" does not exist in {self}')
 
@@ -326,6 +422,7 @@ class Outbox(Manifold):
 
 
 class Filter(Component):
+
     def __call__(self, messages):
         return self.filter(messages)
 
@@ -352,6 +449,34 @@ class ChannelChain(Filter):
                 return test_name, message
 
 
+class MessageMangleFilter(Filter):
+
+    def deepget(self, mapping, key):
+        if '.' not in key:
+            return mapping[key]
+
+        else:
+            current, remainder = key.split('.', 1)
+            return self.deepget(mapping[current], remainder)
+
+    def deepset(self, mapping, key, value):
+        if '.' not in key:
+            mapping[key] = value
+
+        else:
+            current, remainder = key.split('.', 1)
+            self.deepset(mapping[current], remainder, value)
+
+    async def process(self, channel, message):
+        for spec in self.settings.attributes:
+            if len(spec) != 1:
+                raise ValueError(f'Bad configuration: {self.settings}')
+
+            key = list(spec.keys())[0]
+            self.deepset(message, key, spec[key].format(**message))
+        return channel, message
+
+
 class ThreadedNode(Node):
     pass
 
@@ -382,27 +507,59 @@ class ThreadedProcessor:
             if i:  # skip the one already consumed by aiter above
                 await self.oqueue.async_q.get()
             self.oqueue.async_q.task_done()
+            # print(f'thread {i} EOT received back')
 
-        while self.threads:
-            asyncio.sleep(0.1)
+        while self.threads:  # FIXME: join threads instead, and move or retire del self.threads[]
+            asyncio.sleep(0.1)  #     while at it, catch exceptions and clean up queue with task_done  # noqa: E501
+        # print(f'all threads stopped')
 
     async def thread_feeder(self, messages):
         async for (channel, message) in messages:
+            # TODO: recreate dead threads, report thread failures
             await self.iqueue.async_q.put((channel, message))
         for i in range(self.scale):
             await self.iqueue.async_q.put((None, Message.EOT))
+            # print(f'thread {i} EOT sent')
 
     def thread_consumer(self, threadnum):  # runs in thread
+        # print(f'thread {threadnum} started')
         queue_iter = iter(self.iqueue.sync_q.get, (None, Message.EOT))
         for channel, message in self.processor(queue_iter):
             self.iqueue.sync_q.task_done()
             self.oqueue.sync_q.put((channel, message))
+        # print(f'thread {threadnum} EOT received')
         self.iqueue.sync_q.task_done()  # for EOT which does not make it past queue iterator
+        # print(f'thread {threadnum} task done')
         self.barrier.wait()
-        self.oqueue.sync_q.put((None, Message.EOT))
+        self.oqueue.sync_q.put((None, Message.EOT))  # TODO: use threading.Event to send just one
+        # print(f'thread {threadnum} EOT sent back')
         del self.threads[threadnum]
+        # print(f'thread {threadnum} stopped')
 
 
 class PriorityRegistry(dict):
     def ordered(self):
         return [item for _, item in sorted(self.items())]
+
+
+class CmdRunner:
+
+    async def runcmd(self, *args, raise_=True, **kwargs):
+        try:
+            proc = await asyncio.create_subprocess_exec(
+                *map(str, args), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE,
+                **kwargs)
+            await proc.wait()
+            exitcode = await proc.wait()
+            stdout = (await proc.stdout.read()).decode()
+            stderr = (await proc.stderr.read()).decode()
+            if exitcode:
+                raise RuntimeError(f'Command {args} ({kwargs}) returned with exitcode {exitcode}, '
+                                f'stdout: {stdout or None}, stderr: {stderr or None}')
+
+        except Exception as e:
+            self.logger.exception(f'Error running command: {e}')
+            if raise_:
+                raise
+
+        return exitcode, stdout, stderr
