@@ -2,8 +2,8 @@
 
 import asyncio
 import logging
-from datetime import datetime
-from uuid import uuid4
+from datetime import datetime  # maybe use udatetime
+from uuid import uuid4  # maybe use fastuuid (incompatible with PyPy)
 
 from box import Box
 
@@ -11,9 +11,10 @@ from .component import Component
 from .utils import aiter
 
 _l = logging.getLogger(__name__)
+stop = False
 
 
-class Message(Box):
+class Message:
     """Data wrapper."""
 
     _instances = dict()  # TODO: check at end of workflow if any undropped messages are remaining
@@ -28,11 +29,24 @@ class Message(Box):
                    created=datetime.utcnow(),
                    history=list(),
                    dropped=False)
-        super().__init__(data, meta=meta)
-        self._instances[self.meta.id] = self
+        self._instances[meta.id] = self
+        self.__data = Box(data, meta=meta.to_dict(), box_dots=True)
+        setattr(self, f'__setattr__', getattr(self, f'__Xsetattr__'))  # prevent recursion
+
+    def __getattr__(self, name, **kwargs):
+        return getattr(self.__data, name, **kwargs)
+
+    def __Xsetattr__(self, name, value):
+        return setattr(self.__data, name, value)
+
+    def __getitem__(self, name, **kwargs):
+        return self.__data.get(name, **kwargs)
+
+    def __setitem__(self, name, value):
+        self.__data[name] = value
 
     def __str__(self):
-        return f'<{self.__class__.__name__} {self.meta.id}>'
+        return f'<{self.__class__.__name__} {self.get("file") or self.get("path") or list(self)} {self.meta.id}>'
 
     def checkin(self, component):
         self._tag(component, 'in')
@@ -53,6 +67,22 @@ class Message(Box):
     def _tag(self, component, tag):
         self.meta.history.append((datetime.utcnow(), component.id, tag))
 
+    # def deepget(self, mapping, key):
+    #     if '.' not in key:
+    #         return mapping[key]
+    #
+    #     else:
+    #         current, remainder = key.split('.', 1)
+    #         return self.deepget(mapping[current], remainder)
+    #
+    # def deepset(self, mapping, key, value):
+    #     if '.' not in key:
+    #         mapping[key] = value
+    #
+    #     else:
+    #         current, remainder = key.split('.', 1)
+    #         self.deepset(mapping[current], remainder, value)
+
     class __EOT__:
         """Used to signal the end of transmission of a queue."""
 
@@ -97,7 +127,12 @@ class Pipe(Component):
 
     def drop(self, message):
         if isinstance(message, Message):
+          try:
+            # self.debug(f'--- dropping {message}')
             message.drop(self)
+          except:
+            __import__('pudb').set_trace()
+            raise
 
     async def receive(self, **kwargs):
         raise NotImplementedError(f'receive() out {self}')
@@ -136,13 +171,16 @@ class Manifold(Pipe):
     def __str__(self):
         return '<%s [%s]>' % (self.__class__.__name__,
                               ', '.join('%s:%s' % (n, p.id)
+                              # ', '.join('%s:%s' % (n, p.get('id', f'--{p}'))
                                         for n, p in self.channels.items()))
 
     async def run(self):
         await super().run()
         self.debug(f'- starting pipes')
+        # __import__('pudb').set_trace()
         for pipe in self.channels.values():
-            await pipe.start()
+            if not pipe.running:
+                await pipe.start()
         self.ready.set()
 
 
@@ -162,6 +200,7 @@ class DataPipe(Pipe):
     """Outputs preloaded messages, in order."""
 
     def configure(self, messages=None):
+        # import pudb; pudb.set_trace()
         self.messages = [Message(**m) for m in reversed(messages or [])]
         return dict(messages=len(self.messages))
 
@@ -180,6 +219,9 @@ class DataPipe(Pipe):
             except IndexError:
                 break
 
+    def empty(self):
+        return len(self.messages) == 0
+
 
 class QueuePipe(Pipe):
     def configure(self, queue=None, maxsize=1):
